/*  Implements the Gauss-Seidel scheme for solving Poisson's equation
    in a rectangular geometry with periodic boundary conditions. This
    is achieved by letting the "boundary" rows and columns hold copies
    of the last interior row or column on the other side of the
    domain. THese are updated each iteration.

    Optional comand line arguments specify the location of point
    charges to be placed in the domain:

         a.out strength1 x1 y1 strength2 x2 y2 ...

    where 

    * strength is the charge density (can be negative); and

    * x and y are the coordinates specifying its location, in units of
      the grid spacing (i.e., (x,y) = (12,5) puts the charge at grid
      node 12, 5). */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define XPOINTS  98    /* Grid size in x direction (not including bdy
			   points) */
#define YPOINTS  98    /* Grid size in y direction (not including bdy
			   points) */
#define MAXITERS 50000  /* Max number of iterations allowed */
#define GRID     0.1    /* Grid spacing in physical units */
#define TOL      1.e-6  /* Numerical tolerance for convergence */
#define EPS      1.e-12 /* A small parameter */

/* Boundary values (could be command line args) */
/* #define LEFT   5.0 */
/* #define RIGHT  5.0 */
/* #define TOP    10.0 */
/* #define BOTTOM 10.0 */

#ifndef PI
#define PI 4.0*atan(1.0)
#endif


int main (int argc, char *argv[])
{
  /* Here there is no need for a separate array to hold the updated
     values. */
  double V[YPOINTS+2][XPOINTS+2];
  double rho[YPOINTS+2][XPOINTS+2];
  double str, h, h2, change, maxChange, Vold, pih2;
  int i, j, k, iter;
  int ip, jp;
  double Ex, Ey, Exc, Eyc;
  FILE *fp;

  h = GRID; h2 = h*h;
  pih2 = PI*h2;

  /* Temporary use of str */
  str = 350;

  /* Interior points */
  for (i=1; i<XPOINTS+1; i++)
    for (j=1; j<YPOINTS+1; j++) {
      V[j][i] = str;
      rho[j][i] = 0.0;
    }


  /* Process any args specifying charges: */
  if (argc > 1) {

    if ((argc - 1) % 3 != 0) {
      printf("Num args must be multiple of three!\n");
      exit(1);
    }
    k = 1;
    while (argv[k] != NULL) {
      str = atof (argv[k++]);
      j = atoi(argv[k++]);
      i = atoi(argv[k++]);
      rho[i][j] = str;
    }
  }
  
  /* Begin relaxation */
  iter = 0;
  while (iter < MAXITERS) {

    maxChange = 0.0;

    /* Set periodic boundary conditions: */

    /* Horizontal boundary surfaces */
    for (k=0; k<XPOINTS+2; k++) {
      V[0][k] = V[YPOINTS][k];
      V[YPOINTS+1][k] = V[1][k];
    }

    /* Vertical boundary surfaces */
    for (k=1; k<YPOINTS+1; k++) {
      V[k][0] = V[k][XPOINTS];
      V[k][XPOINTS+1] = V[k][1];
    }

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++) {
	/* Retain the old value for testing the change. */
	Vold = V[i][j];
	V[i][j] = 0.25*(V[i+1][j] + V[i-1][j] + V[i][j+1] + V[i][j-1]) 
	  + rho[i][j]*pih2;
	if (fabs(Vold) > EPS) {
	  change = fabs(V[i][j]/Vold-1.0);
	  if (change > maxChange) maxChange = change;
	}
      }

    if (maxChange < TOL) {
      printf("Converged in %d iterations...\n", iter+1);
      break;
    }
    iter++;
  }
  if (iter==MAXITERS)
    printf("Warning! Not converged in %d iterations...\n", iter);

  /* Write the solution to a file: */
  fp = fopen ("outfile", "w");

  for (i=0; i<YPOINTS+2; i++) {
    for (j=0; j<XPOINTS+2; j++)
      fprintf(fp, "% 5.3lf\n", V[i][j]);

    /* Note the extra blank line between the data for each row; this
       allows gnuplot to work out the dimensions of the grid
       automatically. */
    fprintf(fp, "\n");
  }
  fclose (fp);

  /* Check: calculate electric field at a reference point */
  ip = 48; jp = 52;
  Ex = Ey = 0.0;

  for (i=1; i<YPOINTS+1; i++)
    for (j=1; j<XPOINTS+1; j++)
      if (fabs(rho[i][j]) > EPS) {

	printf("Point charge of magnitude %lf found at (%d, %d)\n", 
	       rho[i][j], j, i);

	/* Using str for the distance... */
	str = h*sqrt(pow(i-ip,2)+pow(j-jp,2));
/* 	printf("Distance to reference point is %lf\n", str); */
	Ex += 2.0*rho[i][j]*h2/str*((jp-j)*h/str);
	Ey += 2.0*rho[i][j]*h2/str*((ip-i)*h/str);
      }
  printf("Expected result:\n");
  printf("E = (%lf, %lf)\n", Ex, Ey);

  Exc = -(V[ip][jp+1] - V[ip][jp-1])/(2.0*h);
  Eyc = -(V[ip+1][jp] - V[ip-1][jp])/(2.0*h);
  
  printf("Calculated values:\n");
  printf("E = (%lf, %lf)\n", Exc, Eyc);

  printf("Ratio of x-components is %lf\n", Ex/Exc);
  printf("Ratio of y-components is %lf\n", Ey/Eyc);
  printf("Ratio of magnitudes is %lf\n",
	 sqrt(Ex*Ex + Ey*Ey)/sqrt(Exc*Exc + Eyc*Eyc));


  return 0;
}
