/*  Implements Jacobi relaxation for solving Poisson's equation in a
    rectangular geometry, with fixed values on the boundary
    (Dirichlet).

    Optional comand line arguments specify the location of point
    charges to be placed in the domain:

         a.out strength1 x1 y1 strength2 x2 y2 ...

    where 

    * strength is the charge density (can be negative); and

    * x and y are the coordinates specifying its location, in units of
      the grid spacing (i.e., (x,y) = (12,5) puts the charge at grid
      node 12, 5). 

    (See commented line 76 below for an example of other boundary
    options.)
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define XPOINTS  127    /* Grid size in x direction (not including bdy
			   points) */
#define YPOINTS  127    /* Grid size in y direction (not including bdy
			   points) */
#define MAXITERS 50000  /* Max number of iterations allowed */
#define GRID     0.1    /* Grid spacing in physical units */
#define TOL      1.e-6  /* Numerical tolerance for convergence */
#define EPS      1.e-12 /* A small parameter */

/* Boundary values (could be command line args) */
#define LEFT   5.0
#define RIGHT  5.0
#define TOP    10.0
#define BOTTOM 10.0

#ifndef PI
#define PI 4.0*atan(1.0)
#endif


int main (int argc, char *argv[])
{
  float V[YPOINTS+2][XPOINTS+2];
  float Vnew[YPOINTS+2][XPOINTS+2];
  float rho[YPOINTS+2][XPOINTS+2];
  float str, h, h2, change, maxChange, pih2;
  int i, j, k;
  FILE *fp;

  h = GRID; h2 = h*h;
  pih2 = PI*h2;

  /* Temporary use of str */
  str = 0.25*(LEFT + RIGHT + TOP + BOTTOM);

  /* Interior points */
  for (i=1; i<XPOINTS+1; i++)
    for (j=1; j<YPOINTS+1; j++) {
      V[j][i] = str;
      rho[j][i] = 0.0;
    }

  /* Horizontal boundary surfaces */
  for (k=0; k<XPOINTS+2; k++) {
    V[0][k] = BOTTOM;
    V[YPOINTS+1][k] = TOP;
  }

  /* Vertical boundary surfaces */
  for (k=1; k<YPOINTS+1; k++) {
/*     V[k][0] = 10.0*sin(PI*k/(YPOINTS+1)); */
    V[k][0] = LEFT;
    V[k][XPOINTS+1] = RIGHT;
  }

  /* Process any args specifying charges: */
  if (argc > 1) {

    if ((argc - 1) % 3 != 0) {
      printf("Num args must be multiple of three!\n");
      exit(1);
    }
    k = 1;
    while (argv[k] != NULL) {
      str = atof (argv[k++]);
      j = atoi(argv[k++]);
      i = atoi(argv[k++]);
      rho[i][j] = str;
    }
  }
  
  /* Begin relaxation */
  k = 0;
  while (k < MAXITERS) {

    maxChange = 0.0;

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++) {
	Vnew[i][j] = 0.25*(V[i+1][j] + V[i-1][j] + V[i][j+1] + V[i][j-1]) 
	  + rho[i][j]*pih2;
	if (fabs(V[i][j]) > EPS) {
	  change = fabs(Vnew[i][j]/V[i][j] - 1.0);
	  if (change > maxChange) maxChange = change;
	}
      }

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++)
	V[i][j] = Vnew[i][j];

    if (maxChange < TOL) {
      printf("Converged in %d iterations...\n", k+1);
      break;
    }
    k++;
  }
  if (k==MAXITERS)
    printf("Warning! Not converged in %d iterations...\n", k);

  /* Write the solution to a file: */
  fp = fopen ("outfile", "w");

  for (i=0; i<YPOINTS+2; i++) {
    for (j=0; j<XPOINTS+2; j++)
      fprintf(fp, "% 5.3lf\n", V[i][j]);

    /* Note the extra blank line between the data for each row; this
       allows gnuplot to work out the dimensions of the grid
       automatically. */
    fprintf(fp, "\n");
  }
  fclose (fp);

  return 0;
}
