/* Automated multigrid implementation of Gauss-Seidel relaxation for
   solving Poisson's equation in a rectangular geometry, similar to
   the "Full Multigrid Algorithm".

   The simulation starts at grid level 1 (one interior grid point) and
   proceeds one level up in the grid hierarchy until the finest grid
   is reached (see MAXGRID below).  At each level the code iterates to
   the specified relative accuracy. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Does NOT include boundary points */
#define MAXITERS  50000
#define INITGRID  1
#define MAXGRID   127 /* Should be 2^n - 1 */
#define INITTOL   1.e-6
#define FINALTOL  1.e-6

/* Boundary values (could be cmd line args) */
#define LEFT   5.0
#define RIGHT  5.0
#define TOP    10.0
#define BOTTOM 10.0

int main (int argc, char *argv[])
{
  double V[MAXGRID+2][MAXGRID+2];
  double change, maxChange, Vold;
  int i, j, k;
  FILE *fp;
  int istep, newstep;
  int presson = 'y';
  int time = 0;

  /* Interior points */
  for (i=1; i<MAXGRID+1; i++)
    for (j=1; j<MAXGRID+1; j++)
      V[j][i] = 0.0;

  /* Horizontal boundary surfaces */
  for (k=0; k<MAXGRID+2; k++) {
    V[0][k] = BOTTOM;
    V[MAXGRID+1][k] = TOP;
  }

  /* Vertical boundary surfaces */
  for (k=1; k<MAXGRID+1; k++) {
    V[k][0] = LEFT;
    V[k][MAXGRID+1] = RIGHT;
  }

  /* starting value */
  istep = (MAXGRID + 1)/(INITGRID + 1);

  /* Commence multigrid */

  printf ("Initial relaxation...\n");
  
  while (presson == 'y') {

    /* Commence Relaxation */
    k = 0;
    while (k < MAXITERS) {
      
      maxChange = 0.0;

      for (j=istep; j<MAXGRID+1; j += istep) 
	for (i=istep; i<MAXGRID+1; i += istep) {
	  Vold = V[i][j];
	  V[i][j] = 0.25*(V[i+istep][j] + V[i-istep][j] + 
			  V[i][j+istep] + V[i][j-istep]);
	  change = fabs(V[i][j]/Vold-1.0);
	  if (change > maxChange) maxChange = change;
	}
      
      if (maxChange < FINALTOL) {
	printf("Converged in %d iterations...\n", k+1);
	time += (k+1)*((int) pow((MAXGRID+1)/istep - 1, 2));
	break;
      }
      k++;
    }
    if (istep == 1) break;

    newstep = istep/2;

    /* Prolong to finer grid... */
    for (j=newstep; j<MAXGRID+1; j+=istep)
      for (i=istep; i<MAXGRID+1; i+=istep)
	V[i][j] = 0.5*(V[i][j+newstep] + V[i][j-newstep]);

    for (i=newstep; i<MAXGRID+1; i+=istep)
      for (j=istep; j<MAXGRID+1; j+=istep)
	V[i][j] = 0.5*(V[i+newstep][j] + V[i-newstep][j]);
    
    for (i=newstep; i<MAXGRID+1; i+=istep)
      for (j=newstep; j<MAXGRID+1; j+=istep)
	V[i][j] = 0.25*(V[i+newstep][j] + V[i-newstep][j] + 
			V[i][j+newstep] + V[i][j-newstep]);

    istep = newstep;

  }/* End multigrid */

  printf("Done with multigrid... writing final output.\n");

  fp = fopen ("outfile", "w");

  for (i=0; i<MAXGRID+2; i += istep) {
    for (j=0; j<MAXGRID+2; j += istep)
      fprintf(fp, "% 5.3lf\n", V[i][j]);
    fprintf(fp, "\n");
  }
  fclose (fp);

  printf("time = %d\n", time);

  return 0;
}
