/* Implements multigrid version of Gauss-Seidel relaxation for solving
   Poisson's equation in a rectangular geometry.

   Takes as a command line argument the starting grid level n. The
   number of interior grid points is 2^n-1 (see macros defining
   MINGRID and MAXGRID below).

   At each stage the user specifies a _relaxation target_, which is
   interpreted as follows:

     * If less than one, the target is taken as the desired relative
       accuracy. The code iterates until this accuracy is reached.
     * If greater than one, the target is interpreted as a number of
       iterations to be performed. (If a float is specified it is
       truncated to an int.)

  The user is also prompted after each round of iteration for a change
  in the grid level.  This can be positive (make the grid finer) or
  negative (coarser), and may have magnitude greater than one. If the
  specified transformation would go beyond the max or min grid level,
  the max or min level is used instead.

  _time_ is a running tally of arithmetic operations (iterations times
  number of grid points) is maintained as a measure of computational
  work. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Does NOT include boundary points */
#define MAXITERS  50000
#define MINGRID   1      /* Grid level 1 */
#define MAXGRID   127    /* Grid level 7 */
#define INITTOL   1.e-6
#define FINALTOL  1.e-6

/* Boundary values (could be cmd line args) */
#define LEFT   5.0
#define RIGHT  5.0
#define TOP    10.0
#define BOTTOM 10.0

/* -------------------------------------------------------- */
/* Global variables: */

double V[MAXGRID+2][MAXGRID+2];
int gridlevel, istep, maxlevel, minlevel;
int time;

/* -------------------------------------------------------- */
/* Useful for examining small grids as they undergo multigrid
   transformations. */

void PrintGrid ()
{
  int i, j;

  for (i=MAXGRID+1; i>=0; i -= istep) {
    printf("\n");
    for (j=0; j<MAXGRID+2; j += istep)
      printf("% 5.3lf\t", V[i][j]);
  }
  printf("\n");

  return;
}
/* -------------------------------------------------------- */

void WriteGrid (char *outfile)
{
  int i, j;
  FILE *fp;

  fp = fopen ("outfile", "w");

  for (i=0; i<MAXGRID+2; i += istep) {
    for (j=0; j<MAXGRID+2; j += istep)
      fprintf(fp, "% 5.3lf\n", V[i][j]);
    fprintf(fp, "\n");
  }
  fclose (fp);

  return;
}
/* ------------------------------------------------------------ */
/* Implements the prolongation operation described in the text. */

void MGprolong ()
{
  int newstep = istep/2;
  int i, j;
      
  for (j=newstep; j<MAXGRID+1; j+=istep)
    for (i=istep; i<MAXGRID+1; i+=istep)
      V[i][j] = 0.5*(V[i][j+newstep] + V[i][j-newstep]);
  
  for (i=newstep; i<MAXGRID+1; i+=istep)
    for (j=istep; j<MAXGRID+1; j+=istep)
      V[i][j] = 0.5*(V[i+newstep][j] + V[i-newstep][j]);

  for (i=newstep; i<MAXGRID+1; i+=istep)
    for (j=newstep; j<MAXGRID+1; j+=istep)
      V[i][j] = 0.25*(V[i+newstep][j] + V[i-newstep][j] +
		      V[i][j+newstep] + V[i][j-newstep]);

  istep = newstep;
  gridlevel++;

/*   printf ("After prolongation, new index step is %d\n", istep); */
/*   printf("After prolongation, new V = "); */
/*   PrintGrid (); */
  
  return;
}
/* ----------------------------------------------------------- */
/* Implements the restriction operation described in the text. */

void MGrestrict ()
{
  int newstep = istep*2;
  int i, j;

  for (j=newstep; j<MAXGRID+1; j+=newstep)
    for (i=newstep; i<MAXGRID+1; i+=newstep)
      V[i][j] = 0.25*V[i][j] +
	(V[i][j+istep]+V[i][j-istep]+V[i+istep][j]+V[i-istep][j])/8.0 +
	(V[i+istep][j+istep]+V[i+istep][j-istep]+
	 V[i-istep][j+istep]+V[i-istep][j-istep])/16.0;

  istep = newstep;
  gridlevel--;

/*   printf("After restriction, new V = "); */
/*   PrintGrid (); */

  return;
}
/* -------------------------------------------------------- */
/* Processes a multigrid transformation */

int MGtransform (int nsteps) {

  int check, i;

  check = gridlevel + nsteps;
  if (check < minlevel) {
    nsteps = minlevel - gridlevel;
  }
  else if (check > maxlevel) {
    nsteps = maxlevel - gridlevel;
  }

/*   printf("Current V = "); */
/*   PrintGrid (); */

  if (nsteps > 0)
    for (i=0; i<nsteps; i++)
      MGprolong ();
  else if (nsteps < 0)
    for (i=0; i<(-nsteps); i++)
      MGrestrict ();
  else
    ;

  return 0;
}
/* --------------------------------------------------------- */
/* Relaxes the current grid either for a specified number of
   iterations or to a specified relative accuracy. */

void Relax (double relparam)
{
  double change, Vold;
  double maxChange, tol;
  int i, j;
  int iter = 1, maxiters;

  relparam = fabsl (relparam);

  if (relparam > 0.99999) {
    maxiters = (int) relparam;
    tol = INITTOL;
    printf("Performing %d relaxation steps...\n", maxiters);
  }
  else if (relparam > 0) {
    tol = relparam;
    maxiters = MAXITERS;
    printf("Relaxing to relative accuracy %le...\n", tol);
  }
  else {
    tol = INITTOL;
    maxiters = MAXITERS;
  }

  while (iter <= maxiters) {

    maxChange = 0.0;

    for (j=istep; j<MAXGRID+1; j += istep)
      for (i=istep; i<MAXGRID+1; i += istep) {
	Vold = V[i][j];
	V[i][j] = 0.25*(V[i+istep][j] + V[i-istep][j] +
			V[i][j+istep] + V[i][j-istep]);
	change = fabs(V[i][j]/Vold - 1.0);
	if (change > maxChange) maxChange = change;
      }

    if (maxChange < tol) {
      printf("Converged in %d iteration(s)...\n", iter);
      time += iter*((int) pow((MAXGRID+1)/istep - 1, 2));
      return;
    }
    iter++;
  }
  printf("Relaxation NOT converged in %d iterations! maxChange = %le\n", iter-1, maxChange);
  time += (iter - 1)*pow((MAXGRID+1)/istep - 1, 2);

  return;
}
/* -------------------------------------------------------- */

int main (int argc, char *argv[])
{
  double str;
  int i, j, k;
  int presson = 1;
  double relparam;

  minlevel = (int) log2(MINGRID + 1) + 1.e-12;
  maxlevel = (int) log2(MAXGRID + 1) + 1.e-12;

  /* Default initial grid level is 1 */
  if (argc > 1) {
    gridlevel = atoi (argv[1]);
    if (gridlevel < minlevel || gridlevel > maxlevel) {
      printf("Invalid grid level, must be > 0 and < %d\n", maxlevel+1);
      exit(1);
    }
  }
  else gridlevel = 1;

  /* Interior points */
  str = 0.25*(LEFT + RIGHT + TOP + BOTTOM);
  for (i=1; i<MAXGRID+1; i++)
    for (j=1; j<MAXGRID+1; j++)
      V[j][i] = str;

  /* Horizontal boundary surfaces */
  for (k=0; k<MAXGRID+2; k++) {
    V[0][k] = BOTTOM;
    V[MAXGRID+1][k] = TOP;
  }

  /* Vertical boundary surfaces */
  for (k=1; k<MAXGRID+1; k++) {
    V[k][0] = LEFT;
    V[k][MAXGRID+1] = RIGHT;
  }

  /* Starting value */
  istep = (MAXGRID + 1)/((int) pow(2, gridlevel));

  /* Commence multigrid */
  time = 0;
  while (presson != 0) {

    printf("Current grid level is %d\n", gridlevel);
    printf("Relaxation target? ");
    scanf ("%lf", &relparam);

    Relax (relparam);

    printf ("Grid level change? (0=print results and exit) ");
    scanf ("%d", &presson);

    /* Perform multigrid transformation: */
    MGtransform (presson);
  }
  printf("Done with multigrid... writing final output.\n");
  WriteGrid ("outfile");
  printf("Total time = %d\n", time);

  return 0;
}
