/*  Implements the Gauss-Seidel scheme with overrelaxation to solve
    Poisson's equation in a rectangular geometry with fixed values on
    the boundary (Dirichlet).

    The overrelaxation parameter is a required command line argument.
    Note that 
                        1 < OMEGA < 2

    for overrelaxation. OMEGA=1 corresponds to standard Gauss-Seidel.

    Optional additional comand line arguments specify the location of
    point charges to be placed in the domain:

         a.out strength1 x1 y1 strength2 x2 y2 ...

    where 

    * strength is the charge density (can be negative); and

    * x and y are the coordinates specifying its location, in units of
      the grid spacing (i.e., (x,y) = (12,5) puts the charge at grid
      node 12, 5). 
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define XPOINTS  127    /* Grid size in x direction (not including bdy
			   points) */
#define YPOINTS  127    /* Grid size in y direction (not including bdy
			   points) */
#define MAXITERS 50000  /* Max number of iterations allowed */
#define GRID     0.1    /* Grid spacing in physical units */
#define TOL      1.e-6  /* Numerical tolerance for convergence */
#define EPS      1.e-12 /* A small parameter */

/* Boundary values (could be command line args) */
#define LEFT   5.0
#define RIGHT  5.0
#define TOP    10.0
#define BOTTOM 10.0

#ifndef PI
#define PI 4.0*atan(1.0)
#endif


int main (int argc, char *argv[])
{
  double V[YPOINTS+2][XPOINTS+2];
  double rho[YPOINTS+2][XPOINTS+2];
  double str, h, h2, change, maxChange, Vold, pih2, w, oneminusw;
  int i, j, k;
  FILE *fp;


  if (argc < 2) {
    printf("Error: must specify omega\n");
    exit(3);
  }
  w = atof (argv[1]);
  oneminusw = 1.0 - w;

  printf("omega = %f\n", w);

  h = GRID; h2 = h*h;
  pih2 = PI*h2;

  /* Temporary use of str */
  str = 0.25*(TOP + BOTTOM + LEFT + RIGHT);

  /* Interior points */
  for (i=1; i<XPOINTS+1; i++)
    for (j=1; j<YPOINTS+1; j++) {
      V[j][i] = str;
      rho[j][i] = 0.0;
    }

  /* Horizontal boundary surfaces */
  for (k=0; k<XPOINTS+2; k++) {
    V[0][k] = BOTTOM;
    V[YPOINTS+1][k] = TOP;
  }

  /* Vertical boundary surfaces */
  for (k=1; k<YPOINTS+1; k++) {
    V[k][0] = LEFT;
    V[k][XPOINTS+1] = RIGHT;
  }

  /* Process any args specifying charges: */
  if (argc > 2) {

    if ((argc - 2) % 3 != 0) {
      printf("Num args must be multiple of three!\n");
      exit(1);
    }
    k = 2;
    while (argv[k] != NULL) {
      str = atof (argv[k++]);
      j = atoi(argv[k++]);
      i = atoi(argv[k++]);
      rho[i][j] = str;
      printf ("Placeing charge Q = %f at (%d, %d)\n", str, j, i);

    }
  }
  
  /* Begin relaxation */
  k = 0;
  while (k < MAXITERS) {

    maxChange = 0.0;

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++) {
	/* Retain the old value for testing the change. */
	Vold = V[i][j];
	V[i][j] = w*(0.25*(V[i+1][j] + V[i-1][j] + V[i][j+1] + V[i][j-1]) 
		     + rho[i][j]*pih2) + oneminusw*V[i][j];
	if (fabs(Vold) > EPS) {
	  change = fabs(V[i][j]/Vold-1.0);
	  if (change > maxChange) maxChange = change;
	}
      }

    if (maxChange < TOL) {
      printf("Converged in %d iterations...\n", k+1);
      break;
    }
    k++;
  }
  if (k==MAXITERS)
    printf("Warning! Not converged in %d iterations...\n", k);

  /* Write the solution to a file: */
  fp = fopen ("outfile", "w");

  for (i=0; i<YPOINTS+2; i++) {
    for (j=0; j<XPOINTS+2; j++)
      fprintf(fp, "% 5.3lf\n", V[i][j]);

    /* Note the extra blank line between the data for each row; this
       allows gnuplot to work out the dimensions of the grid
       automatically. */
    fprintf(fp, "\n");
  }
  fclose (fp);

  return 0;
}
