/* Implements Gauss-Seidel relaxation for solving Poisson's equation
   in a rectangular geometry with parallel sheets of charge.  

   Takes as cmd line args the y-coordinates of the two sheets and the
   charge per unit length to be placed there, e.g.,

              a.out y1 chg1 y2 chg2 

   Positions specified should be between 1 and YPOINTS. Slabs extend
   across the domain in the x direction, with periodic boundary
   conditions. In the y direction the BCs are antiperiodic. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Does NOT include boundary points */
#define XPOINTS  6
#define YPOINTS  126
#define MAXITERS 500000
#define GRID     0.1
#define TOL      1.e-5
#define EPS      1.e-10

/* Starting guess (don't make this zero!): */
#define INITPOT   10

#ifndef PI
#define PI 4.0*atan(1.0)
#endif


int main (int argc, char *argv[])
{
  double V[YPOINTS+2][XPOINTS+2];
  double rho[YPOINTS+2][XPOINTS+2];
  double h, h2, pih2, change, maxChange, Vold, q1, q2, str;
  int i, j, k, y1, y2, iter;
  FILE *fp;


  /* Set up interior points */
  for (i=1; i<XPOINTS+1; i++)
    for (j=1; j<YPOINTS+1; j++) {
      V[j][i] = INITPOT;
      rho[j][i] = 0.0;
    }

  /* Process args specifying charge slabs */
  if (argc == 5) {

    y1 = atoi(argv[1]);
    if (y1 < 1 || y1 > YPOINTS) {
      printf("Invalid sheet location specified. Exiting...\n");
      exit(1);
    }
    q1 = atof(argv[2]);
    for (j=1; j<XPOINTS+1; j++)
      rho[y1][j] = q1;

    y2 = atoi(argv[3]);
    if (y2 < 1 || y2 > YPOINTS) {
      printf("Invalid sheet location specified. Exiting...\n");
      exit(1);
    }
    q2 = atof(argv[4]);
    for (j=1; j<XPOINTS+1; j++)
      rho[y2][j] = q2;
  }
  else {
    printf("Must specify location and charge/length for slabs!\n");
    exit(2);
  }

  /* Uncomment for fixed boundary conditions for horizontal
     surfaces: */
/*   for (k=0; k<XPOINTS+2; k++) { */
/*     V[0][k] = 0.0; */
/*     V[YPOINTS+1][k] = 0.0; */
/*   } */

  h = GRID; h2 = h*h;
  pih2 = PI*h2;

  /* Commence Relaxation */
  iter = 0;
  while (iter < MAXITERS) {

    /* Copy edge values for periodic BCs */

    /* Horizontal boundary surfaces -- here impose antiperiodic bcs,
       appropriate for equal and oppositely charged slabs. Comment
       this out to experiment with fixed BCs in this direction. */
    for (k=0; k<XPOINTS+2; k++) {
      V[0][k] = -V[YPOINTS][k];
      V[YPOINTS+1][k] = -V[1][k];
    }

    /* Vertical boundary surfaces -- these are periodic BCs. */
    for (k=0; k<YPOINTS+2; k++) {
      V[k][0] = V[k][XPOINTS];
      V[k][XPOINTS+1] = V[k][1];
    }

    maxChange = 0.0;

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++) {
	Vold = V[i][j];
	V[i][j] = 0.25*(V[i+1][j] + V[i-1][j] + V[i][j+1] + V[i][j-1]) 
	  + rho[i][j]*pih2;
	if (fabs(Vold) > EPS) {
	  change = fabs(V[i][j]/Vold-1.0);
	  if (change > maxChange) maxChange = change;
	}
      }

    if (maxChange < TOL) {
      printf("Converged in %d iterations...\n", iter+1);
      break;
    }
    iter++;
  }
  if (iter==MAXITERS)
    printf("Warning! Not converged in %d iterations...\n", iter);

  /* Write results */
  fp = fopen ("outfile", "w");

  for (i=0; i<YPOINTS+2; i++) {
    for (j=0; j<XPOINTS+2; j++)
      fprintf(fp, "% 5.3lf\n", V[i][j]);
    fprintf(fp, "\n");
  }
  fclose (fp);

  /* Check potential difference between plates */
  printf("Plate 1 has potential %lf\n", V[y1][XPOINTS/2]);
  printf("Plate 2 has potential %lf\n", V[y2][XPOINTS/2]);
  printf("Delta V = %lf\n", fabs(V[y1][XPOINTS/2]-V[y2][XPOINTS/2]));
  printf("Plate separation is %d units\n", y2-y1);

  /* In calculating C, note that q1*h is he charge per unit length
     along the slab: */
  printf("Cap per unit length = %lf, expected %lf\n",
	 fabs(q1*h/(V[y1][XPOINTS/2]-V[y2][XPOINTS/2])),
	 1.0/(4.0*PI*(y2-y1)*h));


  /* Calculate the electric field for testing; should be zero above
     and below,nonzero between. */

  /* E field above the slabs */
/*   str = 0.0; */

/*   for (i=YPOINTS; i>=y2+1; i--) { */
/*     change = -(V[i+1][XPOINTS/2]-V[i-1][XPOINTS/2])/(2.0*h); */
/*     printf("At y = %d, E = %lf\n", i, change); */
/*     str += change; */
/*   } */
/*   printf("Average |E| above = %lf\n\n", fabs(str/(YPOINTS-y2))); */

  /* E field between the slabs */
/*   str = 0.0; */
/*   for (i=y2-1; i>=y1+1; i--) { */
/*     change = -(V[i+1][XPOINTS/2]-V[i-1][XPOINTS/2])/(2.0*h); */
/*     printf("At y = %d, E = %lf\n", i, change); */
/*     str += change; */
/*   } */
/*   printf("Average |E| between = %lf\n\n", fabs(str/(y2-y1-1))); */

  /* E field below the slabs */
/*   str = 0.0; */
/*   for (i=y1-1; i>=1; i--) { */
/*     change = -(V[i+1][XPOINTS/2]-V[i-1][XPOINTS/2])/(2.0*h); */
/*     printf("At y = %d, E = %lf\n", i, change); */
/*     str += change; */
/*   } */
/*   printf("Average |E| below = %lf\n\n", fabs(str/(y1-1))); */

  return 0;
}
