/* Implements Gauss-Seidel relaxation for solving Poisson's equation
   in a rectangular geometry with parallel sheets of charge.  

   The sheets are located on the top and bottom sides of the domain,
   and are held at fixed potential. Periodic boundary conditions are
   used in the x direction to make the sheets effectively infinite. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Does NOT include boundary points */
#define XPOINTS  6
#define YPOINTS  126
#define MAXITERS 500000
#define GRID     0.1
#define TOL      1.e-5
#define EPS      1.e-10

/* Starting boundary values */
#define LEFT   10
#define RIGHT  10
#define TOP    10
#define BOTTOM -10

#ifndef PI
#define PI 4.0*atan(1.0)
#endif


int main (int argc, char *argv[])
{
  double V[YPOINTS+2][XPOINTS+2];
  double rho[YPOINTS+2][XPOINTS+2];
  double str, h, h2, pih2, change, maxChange, Vold, q1, q2, tmp;
  int i, j, k, y1, y2, iter;
  FILE *fp;

  /* Temporary use of str */
  str = 0.25*(TOP+BOTTOM+LEFT+RIGHT);

  /* Set up interior points */
  for (i=1; i<XPOINTS+1; i++)
    for (j=1; j<YPOINTS+1; j++) {
      V[j][i] = str;
      rho[j][i] = 0.0;
    }

  /* Fixed boundary conditions for horizontal surfaces: */
  for (k=0; k<XPOINTS+2; k++) {
    V[0][k] = BOTTOM;
    V[YPOINTS+1][k] = TOP;
  }

  h = GRID; h2 = h*h;
  pih2 = PI*h2;

  /* Commence Relaxation */
  iter = 0;
  while (iter < MAXITERS) {

    /* Copy edge values for periodic BCs */

    /* Vertical boundary surfaces -- these are periodic. */
    for (k=0; k<YPOINTS+2; k++) {
      V[k][0] = V[k][XPOINTS];
      V[k][XPOINTS+1] = V[k][1];
    }

    maxChange = 0.0;

    for (j=1; j<XPOINTS+1; j++) 
      for (i=1; i<YPOINTS+1; i++) {
	Vold = V[i][j];
	V[i][j] = 0.25*(V[i+1][j] + V[i-1][j] + V[i][j+1] + V[i][j-1]) 
	  + PI*rho[i][j]*pih2;
	if (fabs(Vold) > EPS) {
	  change = fabs(V[i][j]/Vold-1.0);
	  if (change > maxChange) maxChange = change;
	}
      }

    if (maxChange < TOL) {
      printf("Converged in %d iterations...\n", iter+1);
      break;
    }
    iter++;
  }
  if (iter==MAXITERS)
    printf("Warning! Not converged in %d iterations...\n", iter);

  /* Write results */
  fp = fopen ("outfile", "w");

  for (i=0; i<YPOINTS+2; i++) {
    for (j=0; j<XPOINTS+2; j++)
      fprintf(fp, "% 5.3lf\n", V[i][j]);
    fprintf(fp, "\n");
  }
  fclose (fp);

  fp = fopen ("charges", "w");

  for (i=1; i<YPOINTS+1; i++) {
    for (j=1; j<XPOINTS+1; j++)
      fprintf(fp, "% 5.3lf\n", rho[i][j]);
    fprintf(fp, "\n");
  }
  fclose (fp);

  y1 = 0; y2 = YPOINTS + 1;

  /* Check potential difference between plates */
  printf("Delta V = %lf\n", fabs(TOP - BOTTOM));
  printf("Plate separation is %d units\n", y2-y1);

  /* Calculate charge per unit length */
  q1 = -(V[y1+2][XPOINTS/2]-V[y1][XPOINTS/2])/(2.0*h*4.0*PI);
  q2 =  (V[y2-2][XPOINTS/2]-V[y2][XPOINTS/2])/(2.0*h*4.0*PI);

  printf("q1 = %lf\n", q1);
  printf("q2 = %lf\n", q2);

  printf("Cap per unit area = %lf, expected %lf\n",
	 fabs(q1/(TOP - BOTTOM)),
	 1.0/(4.0*PI*(y2-y1)*h));

  return 0;
}
